import concurrent.futures
import time
import weakref

from .client import sendmsg, DEFAULT_PORT
from .exceptions import DeadExecutor


class Future(concurrent.futures.Future):
    """Encapsulates the asynchronous execution of a task.
    
    Future instances are created by :py:meth:`apium.TaskExecutor.submit`,
    :py:meth:`~apium.Future.then` and :py:meth:`~apium.Future.catch`, and
    should not be created directly except for testing.
    """

    def __init__(self, task_id, executor, *args, **kwargs):
        self._executor_ref = weakref.ref(executor)
        self._id = task_id
        super(Future, self).__init__(*args, **kwargs)

    def _executor(self):
        executor = self._executor_ref()
        if not executor or not executor._running:
            raise DeadExecutor()
        return executor

    def cancel(self):
        """Cancel the Future if possible.

        Returns True if the future was cancelled, False otherwise. A future
        cannot be cancelled if it is running or has already completed.
        """
        self._executor()._cancel_task(self)
        result = super(Future, self).cancel()
        if result:
            del self._executor()._tasks[self._id]
        return result

    def then(self, fn, *args, **kwargs):
        """Chain a new task to the result of this task.

        Once the task encapsulated by this :py:class:`~apium.Future` finishes,
        the result will be passed as the first argument to the new task. If
        instead of a result, an exception is raised, the exception will
        propagate to the new task (ie. instead of running the task, it will
        simply take the parent task's exception).

        ::

            >>> import apium
            >>> with apium.TaskExecutor() as executor:
            ...     t1 = executor.submit('add', 1, 2)
            ...     t2 = t1.then('add', 3)
            ...     print(t1.result())
            ...     print(t2.result())
            3
            6

        :param task_name: The name of the task to chain.
        :type task_name: str
        :param args: The arguments to pass to the new task. Note that these
            arguments will be passed after and in addition to the result of
            this Future's task.
        :type args: tuple
        :param kwargs: The keyword arguments to pass to the task.
        :type kwargs: dict
        :returns: A Future encapsulating this task
        :rtype: :py:class:`~apium.Future`
        :raises RuntimeError: if the Executor has already shut down.
        :raises TaskDoesNotExist: if a task with the given name doesn't exist.
        """
        return self._executor()._chain(self, "then", fn, *args, **kwargs)

    def catch(self, task_name, *args, **kwargs):
        """Chain a new task to any exception raised by this task.

        Once the task encapsulated by this :py:class:`~apium.Future` finishes,
        any exception raised by this Future's task will be passed as the first
        argument to the new task. If this task doesn't raise an exception, the
        result will propagate to the result of the new task (ie. instead of
        running the task, it will simply take the parent task's result).

        ::

            >>> import apium
            >>> with apium.TaskExecutor() as executor:
            ...     t1 = executor.submit('exc_raise')
            ...     t2 = t1.catch('exc_to_string')
            ...     print(t2.result())
            "Exception('exc_raise')"

        :param task_name: The name of the task to chain.
        :type task_name: str
        :param args: The arguments to pass to the new task. Note that these
            arguments will be passed after and in addition to the exception
            raised by this Future's task.
        :type args: tuple
        :param kwargs: The keyword arguments to pass to the task.
        :type kwargs: dict
        :returns: A Future encapsulating this task
        :rtype: :py:class:`~apium.Future`
        :raises RuntimeError: if the Executor has already shut down.
        :raises TaskDoesNotExist: if a task with the given name doesn't exist.
        """
        return self._executor()._chain(self, "catch", task_name, *args, **kwargs)


class TaskExecutor(concurrent.futures.Executor):
    """An Executor subclass that executes tasks asynchronously using a pool of
    remote workers. The polling_interval dictates how often the TaskExecutor
    should poll the workers for the current state of incomplete tasks.
    """

    def __init__(self, server='localhost', port=DEFAULT_PORT, polling_interval=1):
        self._address = (server, port)
        self._polling_interval = polling_interval
        self._tasks = {}
        self._running = True
        self._shutting_down = False

        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self._executor.submit(self._poll)

    def _update_details(self, future, details):
        try:
            state = details['state']
            notifiable_states = [
                concurrent.futures._base.CANCELLED_AND_NOTIFIED,
                concurrent.futures._base.RUNNING
            ]
            if future._state != state and state in notifiable_states:
                # TODO: Need to delete this task from self._tasks if cancelled.
                future.set_running_or_notify_cancel()
        except KeyError:
            return
        result = details['result']
        if future._result != result:
            future.set_result(result)
            del self._tasks[future._id]
            return
        exc = details['exception']
        if future._exception != exc:
            future.set_exception(exc)
            del self._tasks[future._id]
            return
        future._state = state

    def _poll(self):
        while self._running:
            time.sleep(self._polling_interval)
            for task_id, future in self._tasks.copy().items():
                details = sendmsg(self._address, {'op': 'poll', 'id': task_id})
                self._update_details(future, details)

    def _cancel_task(self, future):
        details = sendmsg(self._address, {'op': 'cancel', 'id': future._id})
        self._update_details(future, details)

    def _new_future(self, details):
        future = Future(details['id'], self)
        self._tasks[details['id']] = future
        return future

    def _chain(self, future, chain_type, fn, *args, **kwargs):
        task = sendmsg(self._address, {
            'op': 'chain',
            'task': {'name': fn, 'args': args, 'kwargs': kwargs, 'type': chain_type},
            'parent': future._id,
        })
        return self._new_future(task)

    def submit(self, task_name, *args, **kwargs):
        """Submits a task to be executed with the given arguments.

        Schedules the task to be executed as fn(\*args, \*\*kwargs) and returns
        a Future instance representing the execution of the remote task.

        :param task_name: The name of the task to run.
        :type task_name: str
        :param args: The arguments to pass to the task.
        :type args: tuple
        :param kwargs: The keyword arguments to pass to the task.
        :type kwargs: dict
        :returns: A Future encapsulating this task
        :rtype: :py:class:`~apium.Future`
        :raises RuntimeError: if the Executor has already shut down.
        :raises TaskDoesNotExist: if a task with the given name doesn't exist.
        """
        if self._shutting_down:
            raise RuntimeError('cannot schedule new tasks after shutdown')
        task = sendmsg(self._address, {'op': 'submit', 'task': {'name': task_name, 'args': args, 'kwargs': kwargs}})
        return self._new_future(task)

    def shutdown(self, wait=True):
        """Clean-up the resources associated with the Executor.

        It is safe to call this method several times. Otherwise, no other
        methods can be called after this one.

        You can avoid having to call this method explicitly if you use the
        `with` statement, which will shutdown the Executor (waiting as if
        shutdown() were called with wait set to True)

        :param wait: If True then shutdown will not return until all running
            futures have finished executing and the resources used by the
            executor have been reclaimed.
        :type wait: bool
        """
        self._shutting_down = True
        if wait:
            while self._tasks:
                time.sleep(self._polling_interval)
        self._running = False



import collections
import concurrent.futures
import contextlib
import functools
import importlib
import logging
import pickle
import threading
import time
import traceback
import uuid

from datetime import datetime
from multiprocessing import Queue, Lock

from .client import sendmsg
from .exceptions import RemoteException, TaskDoesNotExist, UnknownMessage
from .utils import format_fn


try:
    from inspect import signature
except ImportError:
    from funcsigs import signature


try:
    import socketserver
except ImportError:
    import SocketServer as socketserver


tasks = {}
futures = {}
chains = collections.defaultdict(list)
chains_lock = Lock()
schedule_queue = Queue()
schedules = collections.defaultdict(list)


def register_task(fn):
    """A decorator to register a function as a task, to be later run by a
    Worker.

    .. doctest::

        >>> import apium
        >>> @apium.register_task
        ... def task1(arg1, kwarg1=1):
        ...     return arg1 + kwarg1

    :param fn: The function to register.
    :type fn: function
    :returns: The given function.
    :rtype: function
    """
    fn_name = fn.__name__
    logging.debug('Registering task %s', fn_name)
    tasks[fn_name] = fn
    return fn


def schedule_task(start_when=None, repeat_every=None, args=None, kwargs=None):
    """A decorator to register and schedule the given function at a time and/or
    interval.

    It's possible to apply this decorator multiple times to the same function in
    order to have multiple schedules for the same function.

    .. testsetup::

        from datetime import datetime, timedelta

    .. doctest::

        >>> import apium
        >>> @apium.schedule_task(datetime.now() + timedelta(seconds=30), args=(2, ))
        ... @apium.schedule_task(repeat_every=timedelta(seconds=60), args=(3, ), kwargs={'kwarg1': 3})
        ... @apium.schedule_task(args=(4, ), kwargs={'kwarg1': 5})
        ... def task1(arg1, kwarg1=1):
        ...     return arg1 + kwarg1

    In the example above, task1 will be added to the worker's queue:

     - as task1(2, kwarg1=1), 30 seconds after the workers start.

     - as task1(3, kwarg1=3), immediately after the workers start and every 60 seconds thereafter.

     - as task1(4, kwarg1=5), immediately after the workers start.

    :param start_when: When to run the function for the first time. A value of
        None results in the function being run immediately after the workers
        start.
    :type start_when: datetime.datetime
    :param repeat_every: The interval at which the function should repeatedly
        run. A value of None results in the function only being called once.
    :type repeat_every: datetime.timedelta
    :param args: The arguments to pass to the scheduled function.
    :type args: tuple
    :param kwargs: The keyword arguments to pass to the scheduled function.
    :type kwargs: dict
    :returns: The given function
    :rtype: function
    """
    def _schedule_task(fn):
        fn_name = fn.__name__
        logging.debug(
            'Scheduling task %s for %s, repeating every %s',
            format_task({'name': fn_name, 'args': args, 'kwargs': kwargs}),
            start_when or datetime.now(),
            repeat_every
        )
        schedule = (start_when or datetime.now(), repeat_every, args or (), kwargs or {})
        schedule_queue.put((fn_name, schedule))
        return register_task(fn)
    return _schedule_task


def format_task(task, chain=False):
    if chain:
        task = task.copy()
        task['args'] = ('X', ) + task['args']

    return '{}{}{}'.format(
        format_fn(task['name'], task['args'], task['kwargs']),
        ' [{}]'.format(task['id'].decode()) if task.get('id') else '',
        ' from {}'.format(task['client']) if task.get('client') else '',
    )


def run_task(task):
    formatted_task = format_task(task)
    logging.info('Running task %s...', formatted_task)
    task_start = datetime.now()
    try:
        result = tasks[task['name']](*task['args'], **task['kwargs'])
        logging.info(
            'Task %s took %s seconds, and returned %s',
            formatted_task,
            (datetime.now() - task_start).total_seconds(),
            result,
        )
        return result
    except Exception:
        logging.info(
            'Task %s took %s seconds, and raised an exception:\n%s',
            formatted_task,
            (datetime.now() - task_start).total_seconds(),
            traceback.format_exc(),
        )
        remote_exc = RemoteException(traceback.format_exc())
        raise remote_exc


def scheduler_process(address, interval, schedule_queue):
    try:
        while True:
            start_time = datetime.now()
            while not schedule_queue.empty():
                new_schedule = schedule_queue.get(False)
                if new_schedule is None:
                    return
                fn_name, schedule = new_schedule
                schedules[fn_name].append(schedule)
            for fn_name, fn_schedules in schedules.copy().items():
                for loc, schedule in enumerate(fn_schedules[:]):
                    next_run, repeat_every, args, kwargs = schedule
                    if next_run < start_time:
                        if repeat_every:
                            fn_schedules[loc] = (next_run + repeat_every, repeat_every, args, kwargs)
                        else:
                            fn_schedules.pop(loc)
                        sendmsg(address, {'op': 'submit', 'task': {'name': fn_name, 'args': args, 'kwargs': kwargs}})
            sleep_duration = max(interval - (datetime.now() - start_time).total_seconds(), 0.1)
            time.sleep(sleep_duration)
    except KeyboardInterrupt:
        pass


def get_future_state(future):
    if future is None:
        return {}
    with future._condition:
        return {
            'state': future._state,
            'result': future._result,
            'exception': future._exception,
        }


def check_for_chain(parent_id, executor, parent_future):
    with chains_lock:
        for task in chains[parent_id]:
            add_task(executor, task, parent_future)
        del chains[parent_id]


def future_values_fall_through(task_id, parent_future):
    future = concurrent.futures.Future()
    future._state = parent_future._state
    future._result = parent_future._result
    future._exception = parent_future._exception
    futures[task_id] = future


def submit_new_task(executor, task):
    future = executor.submit(run_task, task)
    future.add_done_callback(functools.partial(check_for_chain, task['id'], executor))
    future._task = task
    futures[task['id']] = future


def add_task(executor, task, parent_future):
    try:
        task['args'] = (parent_future.result(), ) + task['args']
    except BaseException as err:
        if task['type'] == 'catch':
            task['args'] = (err, ) + task['args']
            submit_new_task(executor, task)
        else:
            future_values_fall_through(task['id'], parent_future)
    else:
        if task['type'] == 'then':
            submit_new_task(executor, task)
        else:
            future_values_fall_through(task['id'], parent_future)


@contextlib.contextmanager
def create_workers(address, modules, num_workers, interval):

    logging.debug('Setting up process pool with %s workers', num_workers)
    executor = concurrent.futures.ProcessPoolExecutor(max_workers=num_workers)

    for mod in modules:
        logging.debug('Importing %s', mod)
        importlib.import_module(mod)

    logging.debug('Setting up task scheduler thread')
    scheduler = threading.Thread(
        target=scheduler_process,
        args=(address, interval, schedule_queue),
        name='Scheduler',
    )
    scheduler.daemon = True

    class TCPHandler(socketserver.BaseRequestHandler):

        def handle(self):
            self.data = self.request.recv(10240).strip()
            request = pickle.loads(self.data)
            logging.debug('Received request from %s: %s', self.client_address[0], request)
            # TODO: Check credentials
            # TODO: Encrypt credentials
            try:
                if request['op'] == 'submit':
                    # TODO: Use the client address if the client says task is "private".
                    task = request['task']
                    if task['name'] not in tasks:
                        self.request.sendall(pickle.dumps(TaskDoesNotExist(task['name'])))
                        return
                    task['client'] = self.client_address[0]
                    task_id = str(uuid.uuid4()).encode()
                    task['id'] = task_id
                    logging.debug('Adding task %s to queue', format_task(task))
                    submit_new_task(executor, task)
                    self.request.sendall(pickle.dumps(task))
                elif request['op'] == 'cancel':
                    task_id = request['id']
                    future = futures[task_id]
                    if future is not None:
                        response = future.cancel()
                    else:
                        response = False
                        with chains_lock:
                            for children in chains.values():
                                for child in children[:]:
                                    if child['id'] == task_id:
                                        response = True
                                        children.remove(child)
                    if response:
                        logging.debug('Cancelled task [%s]', task_id.decode())
                    else:
                        logging.debug('Failed to cancel task [%s]', task_id.decode())
                    state = get_future_state(future)
                    state['response'] = response
                    self.request.sendall(pickle.dumps(state))
                elif request['op'] == 'poll':
                    task_id = request['id']
                    future = futures[task_id]
                    state = get_future_state(future)
                    logging.debug('Responding to poll for task [%s] with: %s', task_id.decode(), state)
                    self.request.sendall(pickle.dumps(state))
                elif request['op'] == 'chain':
                    # TODO: Use the client address if the client says task is "private".
                    task = request['task']
                    parent_id = request['parent']
                    if task['name'] not in tasks:
                        self.request.sendall(pickle.dumps(TaskDoesNotExist(task['name'])))
                        return
                    task['client'] = self.client_address[0]
                    task_id = str(uuid.uuid4()).encode()
                    task['id'] = task_id
                    future = futures[parent_id]
                    if future is not None and future.done():
                        add_task(executor, task, future)
                    else:
                        with chains_lock:
                            chains[parent_id].append(task)
                        futures[task_id] = None
                    logging.debug('Chaining task %s to [%s]', format_task(task, chain=True), parent_id.decode())
                    self.request.sendall(pickle.dumps(task))
                elif request['op'] == 'inspect':
                    task_list = {}
                    for task_name, task_fn in tasks.items():
                        try:
                            sig = signature(task_fn)
                        except ValueError as err:
                            sig = err
                        task_list[task_name] = str(sig)
                    response = {
                        'tasks': task_list,
                        'schedules': schedules,
                        'running': [f._task for f in futures.values() if f and f.running()],
                    }
                    self.request.sendall(pickle.dumps(response))
                else:
                    self.request.sendall(pickle.dumps(UnknownMessage(request)))
            except:
                logging.warning('Exception raised when processing request:\n%s', traceback.format_exc())
                self.request.sendall(pickle.dumps(UnknownMessage(request)))

    logging.debug('Setting up TCP server')
    socketserver.ThreadingTCPServer.allow_reuse_address = True
    server = socketserver.ThreadingTCPServer(address, TCPHandler)

    logging.debug('Starting task scheduler')
    scheduler.start()
    try:
        yield server
    except KeyboardInterrupt:
        print('KeyboardInterrupt')
    finally:
        logging.debug('Stopping task scheduler')
        schedule_queue.put(None)
        scheduler.join()
        logging.debug('Shutting down TCP server')
        server.shutdown()
        server.server_close()
        logging.debug('Shutting down process pool and waiting for currently running tasks to finish')
        executor.shutdown(wait=True)



from .client import sendmsg
from .utils import format_fn


def inspect_worker(address):
    return sendmsg(address, {'op': 'inspect'})


def print_inspected_worker(details):
    if details['tasks']:
        print('Registered tasks:')
        for task_name, task_sig in details['tasks'].items():
            print('  * {}{}'.format(task_name, task_sig))
        print()

    if details['running']:
        print('Running tasks:')
        for task in details['running']:
            print('  * {}:   {}'.format(task['id'].decode(), format_fn(task['name'], task['args'], task['kwargs'])))
        print()

    if details['schedules']:
        single_sched = {t: s for t, s in details['schedules'].items() if [sch for sch in s if not sch[1]]}
        repeat_sched = {t: s for t, s in details['schedules'].items() if [sch for sch in s if sch[1]]}

        if single_sched:
            print('Scheduled tasks:')
            for task_name, schedules in single_sched.items():
                for s in schedules:
                    if not s[1]:
                        print('  * {}:\tScheduled for {}'.format(
                            format_fn(task_name, s[2], s[3]),
                            s[0].isoformat(' '),
                        ))
            print()

        if repeat_sched:
            print('Recurring tasks:')
            for task_name, schedules in repeat_sched.items():
                for s in schedules:
                    if s[1]:
                        print('  * {}:\tScheduled for {}; repeating every {}'.format(
                            format_fn(task_name, s[2], s[3]),
                            s[0].isoformat(' '),
                            str(s[1]),
                        ))
            print()



import argparse
import logging
import multiprocessing

from . import frameworks, worker
from .client import DEFAULT_PORT
from .inspect import print_inspected_worker, inspect_worker


def setup_logging(args):
    FORMAT = '[%(asctime)-15s: %(levelname)+8s/%(processName)+11s]  %(message)s'
    log_kwargs = {
        'format': FORMAT,
        'level': logging.DEBUG if args.debug else logging.INFO
    }
    if args.logfile:
        log_kwargs['filename'] = args.logfile
    logging.basicConfig(**log_kwargs)


def start_workers():
    parser = argparse.ArgumentParser(description='Start workers to run the given tasks.')
    parser.add_argument('modules', nargs='+', help='Modules containing registered tasks.')

    def num_workers_type(x):
        x = int(x)
        if x <= 0:
            raise argparse.ArgumentTypeError("The number of workers must be greater than 0")
        return x

    DEFAULT_NUM_WORKERS = multiprocessing.cpu_count()
    parser.add_argument(
        '-n', '--num-workers',
        dest='num_workers', type=num_workers_type, default=DEFAULT_NUM_WORKERS,
        help='The number of worker processes to start (default {})'.format(DEFAULT_NUM_WORKERS),
    )

    DEFAULT_BIND = 'localhost:{}'.format(DEFAULT_PORT)
    parser.add_argument(
        '-b', '--bind',
        dest='bind', default=DEFAULT_BIND,
        help='The address and port to bind the TCP server to (default {})'.format(DEFAULT_BIND),
    )

    DEFAULT_INTERVAL = 1
    parser.add_argument(
        '-i', '--interval',
        dest='interval', type=float, default=DEFAULT_INTERVAL,
        help='How often the scheduler polls for scheduled tasks in seconds (default {})'.format(DEFAULT_INTERVAL),
    )

    parser.add_argument(
        '-l', '--logfile',
        dest='logfile', default=None,
        help='A filename to write the logs to (default stdout)',
    )

    parser.add_argument(
        '-u', '--username',
        dest='username', default='',
        help='The username to use for authentication (default "")',
    )

    parser.add_argument(
        '-p', '--password',
        dest='password', default='',
        help='The password to use for authentication (default "")',
    )

    parser.add_argument(
        '--debug',
        dest='debug', action='store_const', default=False, const=True,
        help='Show debug logs',
    )

    args = parser.parse_args()
    setup_logging(args)
    server, port = args.bind.rsplit(':', 1)
    address = (server, int(port))

    frameworks.setup()

    with worker.create_workers(address, args.modules, args.num_workers, args.interval) as workers:
        logging.debug('Starting TCP server')
        workers.serve_forever()


def inspect():
    parser = argparse.ArgumentParser(description='Inspect the task queue of the given workers.')

    DEFAULT_BIND = 'localhost:{}'.format(DEFAULT_PORT)
    parser.add_argument(
        '-c', '--connect',
        dest='connect', default=DEFAULT_BIND,
        help='The address and port to bind the TCP server to (default {})'.format(DEFAULT_BIND),
    )

    parser.add_argument(
        '-u', '--username',
        dest='username', default='',
        help='The username to use for authentication (default "")',
    )

    parser.add_argument(
        '-p', '--password',
        dest='password', default='',
        help='The password to use for authentication (default "")',
    )

    args = parser.parse_args()
    server, port = args.connect.rsplit(':', 1)
    address = (server, int(port))

    print_inspected_worker(inspect_worker(address))



from concurrent.futures import (
    CancelledError,
    TimeoutError,
)


class RemoteException(Exception):
    """Raised in place of an Exception raised by the remote task, and contains a
    string representation of the original Exception and its stack trace.
    """
    pass


class TaskDoesNotExist(Exception):
    """Raised when the given task name has not been registered with the remote
    worker.
    """
    pass


class UnknownMessage(Exception):
    """Raised by a worker when it receives a message it can't process."""
    pass


class DeadExecutor(Exception):
    """Raised when a Future's TaskExecutor has shut down, and a method is called
    on a Future that requires communication with the remote worker (eg.
    :py:meth:`~apium.Future.cancel`).
    """
    pass



from concurrent.futures import (
    wait,
    as_completed,
    FIRST_COMPLETED,
    FIRST_EXCEPTION,
    ALL_COMPLETED,
)
from .exceptions import *
from .executor import TaskExecutor, Future
from .worker import register_task, schedule_task

# TODO:
# - Authentication
# - Encryption
# - Private tasks
# - More error checking:
#   - Problems unpickling task on the server.
#   - Problems unpickling results on the client.
#   - Chaining task to non existent parent.
# - Possible to backport a better chaining implementation to concurrent.futures??
#   - Probably not. Would need a reference to the executor in the future.



import pickle
import socket


DEFAULT_PORT = 9737


def sendmsg(address, data):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect(address)
        sock.sendall(pickle.dumps(data))
        received = sock.recv(10240)
        result = pickle.loads(received)
        if isinstance(result, Exception):
            raise result
        return result
    finally:
        sock.close()



import base64
import pickle

from .dh import DiffieHellman


class NoEncryptionLibrariesInstalled(Exception):
    pass


class Encryptor:

    def __init__(self, remote_pk=None):
        self._fern = None
        self._dh = DiffieHellman()
        self._pk = self._dh.public_key
        self._remote_pk = remote_pk

    def _fernet(self):
        if self._fern is None:
            try:
                from cryptography.fernet import Fernet
                self._dh.generate_key(self._remote_pk)
                self._fern = Fernet(base64.urlsafe_b64encode(self._dh.get_key()[:32]))
            except ImportError:
                self._pk = None
                self._fern = 0
        return self._fern

    def encrypt(self, msg):
        return {
            'data': self._fernet().encrypt(pickle.dumps(msg)) if self._fernet() else msg,
            'pk': self._pk,
        }

    def decrypt(self, msg):
        remote_pk = msg['pk']
        if remote_pk:
            self._remote_pk = remote_pk
            if not self._fernet():
                raise NoEncryptionLibrariesInstalled('"cryptography" not installed - cannot decrypt message')
            return pickle.loads(self._fernet().decrypt(msg['data']))
        else:
            return msg['data']



def format_fn(name, args, kwargs):

    def format_args():
        return ', '.join([repr(a) for a in (args or ())])

    def format_kwargs():
        return ', '.join([str(k) + '=' + repr(v) for k, v in (kwargs or {}).items()])

    return '{}({}{}{})'.format(
        name,
        format_args(),
        ', ' if args and kwargs else '',
        format_kwargs(),
    )



from setuptools import setup, find_packages


setup(
    name='apium',
    version='0.0.3',
    author='Christian Boelsen',
    author_email='christianboelsen+github@gmail.com',
    packages=find_packages(exclude=["tests"]),
    entry_points={
        'console_scripts': [
            'apium-worker = apium.command:start_workers',
            'apium-inspect = apium.command:inspect',
        ],
    },
    license='LICENSE',
    description=open('README').read(),
    install_requires=[],
    keywords=['celery', 'task', 'queue'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
    ],
)



import logging
import os


DJANGO_SETTINGS_ENV_VAR = 'DJANGO_SETTINGS_MODULE'


def setup_django():
    try:
        import django
        django.setup()
        from django.core.checks import run_checks
        run_checks()
    except ImportError as error:
        logging.error(
            'Found {} in environment, but couldn\'t import django: {}'.format(DJANGO_SETTINGS_ENV_VAR, error)
        )
        raise


def setup():
    if DJANGO_SETTINGS_ENV_VAR in os.environ:
        setup_django()
